#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import time
import graphlearning as gl
from scipy.optimize import minimize, LinearConstraint
from .graph_utils import degrees
from energy import penergy, jacobian
from numpy.linalg import norm

def grad_descent(gradient, start, rate, tolerance = 1e-2, max_steps = int(1e9)):
    x = start
    for _ in range(max_steps):
        diff = - rate * gradient(x)

        if norm(diff) < tolerance:
            return x
        x += diff

    return x

class ppoisson():
    def __init__(self, p, W):
        self.p = p
        self.W = W
        self.n = W.shape[0]
        self.u = None
        self.fitted = False
        self.predicted = False
        self.runtime = 0
    
    # train_ind: Indices of (few) labeled points
    # train_labels: labels of labaled points (one hot encoding)
    # start: starting point (n,k) array
    def fit(self, train_ind, train_labels, start = np.zeros(1), method = 'trust-constr'):
        if self.fitted:
            return self.u
        
        start_time = time.time()

        self.k = train_labels.shape[1]
        d = degrees(self.W)
        eye = np.eye(self.k)


        if np.count_nonzero(start) == 0:
            model = gl.ssl.poisson(self.W, solver='gradient_descent')
            integer_coded_train_labels = np.argmax(train_labels, axis = 1)

            start = model.fit(train_ind, integer_coded_train_labels) # model's fit 
                                                                     # doesn't expect one hot encoding
        
        if method == 'trust-constr':
            constrain_matrix = np.concatenate([d[i] * eye for i in range(self.n)], axis = 1)
            linear_constraint = LinearConstraint(constrain_matrix, np.zeros(self.k), np.zeros(self.k))
            
            start = start.flatten()
            res = minimize(penergy, x0 = start, args = (self.W, train_ind, train_labels, self.p), 
                    jac = jacobian, method = 'trust-constr', constraints = linear_constraint)
            
            self.u = res.x.reshape(self.n,self.k)
            self.fitted = True
            
            end_time = time.time()
            self.runtime = (end_time - start_time)/60 # in minutes
    
        elif method == 'grad-desc':
            self.fitted = True
            
            self.u = grad_descent(lambda u: jacobian(u, self.W, train_ind, train_labels, self.p), 
                    start=start, rate = 0.5, tolerance = 1e-2)
            end_time = time.time()
            self.runtime = (end_time - start_time)/60 # in minutes

        return self.u

        
    def predict(self):
        if not self.fitted:
            print("Not fitted yet")
            return -1
        if self.predicted:
            return self.predictions
        
        self.predictions = np.argmax(self.u, axis = 1)
        self.predicted = True

        return self.predictions

    
    def fit_predict(self, train_ind, train_labels, start = np.zeros(1)):
        self.fit(train_ind, train_labels, start)
        
        return self.predict()

    # labels: integer valued
    def accuracy(self, labels):
        if not self.predicted:
            self.predictions = self.predict()
        
        return 1 - np.count_nonzero(labels - self.predictions)/self.n

    def print_info(self):
        info_str = f"########### Gradient Descent (w/ Jacobian) for p = {self.p}\n"\
                        f"\nAccuracy = {self.accuracy() * 100:.2f}%\n"\
                        f"Runtime = {self.runtime:.2f} min"

        print(info_str)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.linalg import norm
from numpy.random import rand
from .graph_utils import graph_grad, graph_div
from .pd_alg import power_iteration, pd_alg

def proxF(V, sigma):
    norms = norm(V, axis = 2)

    Z = np.copy(V)
    Z[norms <= sigma] = 0
    Z[norms > sigma] = ((norms[norms > sigma] - sigma)/norms[norms > sigma])[:, np.newaxis] * Z[norms > sigma]
    
    return Z

def proxG(u, tau, mu):
    return tau * mu + u

def wass_dist(W, mu, u0, V0, tol = 1e-2, max_iter = 1e5, verbose = False, rate = 1):
    n = W.shape[0]
    
    K = lambda u: graph_grad(u, W)
    Kstar = lambda V: -graph_div(V, W)
    Kstar_K = lambda u: Kstar(K(u))
    
    _, eig = power_iteration(Kstar_K, rand(n), tol = 1e-3, verbose = False)
    K_norm = np.sqrt(eig)
    sigma = rate/ (2 * K_norm)
    tau = rate/(2 * K_norm )
    
    _proxF = lambda V: proxF(V, sigma)
    _proxG = lambda u: proxG(u, tau, mu)
    
    return pd_alg(_proxF, _proxG, K, Kstar, tau, sigma, u0, V0, 
                  tol = tol, max_iter = max_iter, verbose = verbose)
    



import numpy as np
import pickle
import graphlearning as gl
import sklearn.datasets as datasets
from pyvis.network import Network
from PIL import Image

def save_models(models, filename):
    with open('data_dumps/' + filename, 'wb') as f:
        pickle.dump(models, f)
        f.close()
    
    return

def load_models(filename):
    with open('data_dumps/' + filename, "rb") as f:
        models = pickle.load(f)
        f.close()
    
    return models


"""
Returns the ith vector of the usual basis of R^k
Index starts at 0
"""
def euclidean_basis(i, k):
    eye = np.eye(k)
    return eye[i]

# Example
# print(euclidean_basis(0, 4))
# print(euclidean_basis(2, 4))

# ############## Toy example using gradient descent
# y = np.array([[1,0], [0,1]])
# W = np.array([[0, 1, 1], [1,0,1], [1,1,0]])
# idx = [0, 1]
# p = 2

# u = gradient_ppoisson(W, idx, y, p)
# print("Minimizer:\n", u)
# print("Energy of minimizer:", penergy(u, W, idx, y, p))
# print("Negative Laplacian values:", 2*u[0]-u[1]-u[2], 2*u[1]-u[0]-u[2], 2*u[2]-u[0]-u[1], sep="\n")
# print("Labels:", predict(u))

# ############## Toy example using GraphLearn
# y = np.array([[1,0], [0,1]])
# W = np.array([[0, 1, 1], [1,0,1], [1,1,0]])
# d = degrees(W)
# idx = [0, 1]
# n = 3
# k = 2
# p = 2

# model = gl.ssl.poisson(W)
# u = model.fit(idx, np.array([0, 1]))
# # print(2*u[0]-u[1]-u[2])
# # print(2*u[1]-u[0]-u[2])
# # print(2*u[2]-u[0]-u[1])
# # print(u[0] + u[1] + u[2])
# print("GraphLearn solution: \n", u)

# # Toy example using custom implementation
# my_u = gradient_ppoisson(W, idx, y, p)
# print("Custom solution: \n", my_u)

def construct_data(n, k = 2, labels_per_class=5):
    # Generate training data and label sets
    X,labels = datasets.make_moons(n_samples=n,noise=0.1, random_state = 0)
    W = gl.weightmatrix.knn(X,10).toarray()

    train_ind = gl.trainsets.generate(labels, rate=labels_per_class, seed = 0)
    train_labels = labels[train_ind]
    m = train_ind.size

    # Construction of measure mu (onehot encoded labels)
    # mu[i] = [0, 0] if i is  not in the labeld dataset
    # mu[i] = [1,0] - (mean of train_labels) if i is in the labeled dataset and its label is 1
    # mu[i] = [0,1] - (mean of train_labels) if i is in the labeled dataset and its label is 0
    
    mean = (labels_per_class/m) * np.ones(k)

    a_idx = train_ind[np.argwhere(labels[train_ind] == 1)].flatten()
    b_idx = train_ind[np.argwhere(labels[train_ind] == 0)].flatten()
    mu = np.zeros((n,k))
    mu[a_idx] = [1,0] - mean
    mu[b_idx] = [0,1] - mean

    return X, labels, train_ind, W, mu


# train_labels should be one-hot encoded
def construct_mu(n, k, train_ind, train_labels):
    mu = np.zeros((n,k))
    
    mu[train_ind] = train_labels - train_labels.sum(axis = 0)/train_ind.size

    return mu

def create_graph(X, graph_size = 100, file_name = "example.html", max_width = 40):
    resized_X = np.transpose(X[:graph_size].reshape(X[:graph_size].shape[0], 28, 28), (0, 2, 1))
    reweighted_W = W[:graph_size, :graph_size]
    reweighted_W = max_width * (1/reweighted_W.max()) * reweighted_W

    net = Network()

    for i in range(resized_X.shape[0]):
        img = Image.fromarray(resized_X[i] * 255)
        img = img.convert('RGB')
        img.save(f"data/{i}.png")
        net.add_node(i, shape = 'image', image = f'data/{i}.png')

    for i in range(resized_X.shape[0]):
        for j in range(resized_X.shape[0]):
            if i != j and W[i,j] > 0:
                net.add_edge(i, j, width = reweighted_W[i,j])

    net.toggle_physics(False)
    net.show(file_name)
    
    return 0




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.linalg import norm

# Calculates vector of degrees for the weight matrix W
# W: (n,k) numpy array
def degrees(W):
    return W.sum(axis = 1)

# Example
# Calculates degrees of equilateral triangle
# W = np.array([[0, 1, 1], [1,0,1], [1,1,0]])
# print(degrees(W))

def graph_div(V, W):
    n = V.shape[0]
    k = V.shape[2]
    div = ( V * (W[:, :, np.newaxis] + np.zeros((n,n,k))) ).sum(axis = 1)
    return div

"""
Computes the weighted gradient of u
u: (n,k) matrix
W: (n,n) matrix

Returns:
gradu: (n, n, k)

where gradu[i, j, l] = W[i,j] * u[j,l]-u[i,l]
"""
def graph_grad(u, W):
    gradu = W[:, :, np.newaxis] * (-u[:, np.newaxis] + u)

    return gradu

"""
Computes the non-weighted gradient of u
u: (n,k) matrix

Returns:
gradu: (n, n, k)

where gradu[i, j, l] = u[j,l]-u[i,l]
"""

def grad(u):
    gradu = -u[:, np.newaxis] + u

    return gradu

# # Example:
# u = np.arange(6).reshape(3,2)
# print("u is\n", u)
# print("\nAnd gradu is")
# for i in range(u.shape[0]):
#     for j in range(u.shape[1]):
#         print(graph_grad(u)[i,j], end = "\t\t")
#     print()

# Returns the labels that u predicts
def predict(u):
    return np.argmax(u, axis = 1)

# X: (n,) or (n,k) matrix
# Returns an (n,n)-matrix A whose non-diagonal values are:
#   A[i,j] = 1/norm(X[i] - X[j])
# and whose diagonal values are A[i,i] = 1
def construct_weightmatrix(X): 
    n = X.shape[0]
    
    #diff[i,j] = X[i] - X[j]
    diff = X[:, np.newaxis] - X
    
    
    if X.ndim == 1:
        normed_diff = np.abs(diff)
    else:
        normed_diff = np.apply_along_axis(norm, 2, diff)
    
    normed_diff[np.diag_indices(n)] = np.ones(n)
    
    return 1/normed_diff

# =============================================================================
# Calculates the supremum norm of the graph_grad of u
# u: (n,k)-array
# W: (n,n)-array
# =============================================================================
def sup_graph_grad(u, W):
    print(np.apply_along_axis(norm, 2, graph_grad(u, W)).max())


from setuptools import setup

setup(name='ot_class',
      version='0.2',
      description='Optimal transport methods for classification',
      url='https://github.com/gstamatiou/optimal-classification',
      author='Giorgos Stamatiou',
      author_email='s6gestam@uni-bonn.de',
      license='MIT',
      packages=['ot_class'],
      install_requires=[
          'graphlearning',
          'numpy',
          'matplotlib',
          'Pillow'
          ],
      zip_safe=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy.linalg import norm

def power_iteration(K, x_0, tol = 1e-3, num_iterations = 10000000, verbose = False):
    b = x_0/norm(x_0)
    for _ in range(num_iterations):
        bb = K(b)/norm(K(b))
        diff = norm(b - bb)
        
        if verbose:
            print("PowerIteration:", diff)
            
        if diff < tol:
            return bb, norm(K(bb))/norm(bb)
        b = np.copy(bb)

    return bb, norm(K(bb))/norm(bb)

def pd_alg(proxF, proxG, K, Kstar, tau, sigma, x0, y0, tol = 1e-2, 
           max_iter = 1e6, verbose = False):
    y = y0
    x = x0
    x_bar = x0
    
    difference = tol + 1
    it = 0
    while (it <= max_iter and difference > tol):
        yy = proxF(y + sigma * K(x_bar))
        xx = proxG(x - tau * Kstar(yy))
        xx_bar = 2 * xx - x

        difference = np.max(np.abs(xx - x))
        it += 1

        x = xx
        x_bar = xx_bar
        y = yy

        if verbose:
            print(difference)
    
    if it >= max_iter:
        print("reached max iterations with diff:", difference)
    return x
